__all__ = ["Yaap"]

import sys
import json
import yaml
import warnings
import argparse
from dataclasses import dataclass
from typing import Dict, Sequence, Any, Optional

import jsonschema

from .argument import *
from .utils import private_field


@dataclass
class Yaap:
    name: str = None
    desc: str = None
    on_extra: str = "error"
    add_config_arg: bool = True
    config_arg_name: str = "config"
    add_schema_arg: bool = True
    schema_arg_name: str = "schema"
    _parser: argparse.ArgumentParser = private_field()
    _reserved_parser: argparse.ArgumentParser = private_field()
    _args: Dict[str, Argument] = private_field(default_factory=dict)

    def __post_init__(self):
        if self.on_extra not in {"error", "warn", "ignore"}:
            raise ValueError(f"unrecognized on_extra mode: {self.on_extra}")
        self._parser = argparse.ArgumentParser(
            prog=None if self.name is None else self.name,
            description=None if self.desc is None else self.desc
        )
        self._reserved_parser = argparse.ArgumentParser()
        if self.add_config_arg:
            config_arg = Path(
                name=self.config_arg_name,
                must_exist=True,
                help="Path to a configuration (json/yaml) file "
                     "that contains argument key-values. For keys "
                     "that are specified in both command-line "
                     "arguments and config arguments, the former "
                     "will take precedence."
            )
            self._add_reserved(config_arg)
        if self.add_schema_arg:
            schema_arg = Bool(
                self.schema_arg_name,
                help="Prints the argument schema in json format."
            )
            self._add_reserved(schema_arg)

    @property
    def reserved_arg_names(self):
        return self.config_arg_name, self.schema_arg_name

    def _add_reserved(self, arg: Argument):
        self._args[arg.name] = arg
        args, kwargs = arg.generate_args()
        self._parser.add_argument(*args, **kwargs)
        self._reserved_parser.add_argument(*args, **kwargs)

    def add(self, arg: Argument):
        if arg.name in self._args:
            raise NameError(f"argument of the same name exists: "
                            f"{self._args[arg.name]}")
        if arg.name in self.reserved_arg_names:
            raise NameError(f"argument name is reserved: {arg.name}")
        self._args[arg.name] = arg
        args, kwargs = arg.generate_args()
        self._parser.add_argument(*args, **kwargs)

    def add_int(self,
                name: str,
                shortcut: str = None,
                help: str = None,
                required: bool = False,
                default: Any = None,
                min_bound: int = None,
                max_bound: int = None,
                multiples: int = None,
                choices: Sequence[int] = None,
                is_list: bool = False,
                num_elements: int = None):
        """Specifies an integer-type argument.

        Arguments:
            name (str): Name of the argument. Unlike `argparse`, preceding
                hyphens should be omitted. Allows specification by `--{name}`.
            shortcut (optional, str): Argument shortcut. Allows specification
                by `-{shortcut}`.
            help (optional, str): Help text.
            required (optional, str): Whether this argument is required.
            default (optional, *): Default value for the argument. Must be
                hashable (`__hash__`).
            min_bound (optional, int): Minimum bounds for the integer argument.
                Bound inclusive.
            max_bound (optional, int): Maximum bounds for the integer argument.
                Bound inclusive.
            multiples (optional, int): Enforces the argument to be a multiple
                of this number.
            choices (optional, Sequence[int]): Enforces the argument to be
                one of these candidate numbers.
            is_list (bool): Whether this argument is of list type. If true,
                this argument can be specified with multiple values, i.e.

                    --{name} {val1} {val2} ... {valN}

                (default: False)
            num_elements (optional, int): Number of required elements.
                Applicable only if `is_list` is true.
        """
        kwargs = dict(
            name=name,
            shortcut=shortcut,
            help=help,
            required=required,
            default=default,
            min_bound=min_bound,
            max_bound=max_bound,
            multiples=multiples,
            choices=choices
        )
        if is_list:
            kwargs.update(dict(num_elements=num_elements))
        cls = Int if not is_list else IntList
        return self.add(cls(**kwargs))

    def add_flt(self,
                name: str,
                shortcut: str = None,
                help: str = None,
                required: bool = False,
                default: Any = None,
                min_bound: float = None,
                max_bound: float = None,
                multiples: float = None,
                choices: Sequence[float] = None,
                is_list: bool = False,
                num_elements: int = None):
        """Specifies a float-type argument.

        Arguments:
            name (str): Name of the argument. Unlike `argparse`, preceding
                hyphens should be omitted. Allows specification by `--{name}`.
            shortcut (optional, str): Argument shortcut. Allows specification
                by `-{shortcut}`.
            help (optional, str): Help text.
            required (optional, str): Whether this argument is required.
            default (optional, *): Default value for the argument. Must be
                hashable (`__hash__`).
            min_bound (optional, float): Minimum bounds for the float argument.
                Bound inclusive.
            max_bound (optional, float): Maximum bounds for the float argument.
                Bound inclusive.
            multiples (optional, float): Enforces the argument to be a multiple
                of this number.
            choices (optional, Sequence[float]): Enforces the argument to be
                one of these candidate numbers.
            is_list (bool): Whether this argument is of list type. If true,
                this argument can be specified with multiple values, i.e.

                    --{name} {val1} {val2} ... {valN}

                (default: False)
            num_elements (optional, int): Number of required elements.
                Applicable only if `is_list` is true.
        """
        kwargs = dict(
            name=name,
            shortcut=shortcut,
            help=help,
            required=required,
            default=default,
            min_bound=min_bound,
            max_bound=max_bound,
            multiples=multiples,
            choices=choices
        )
        if is_list:
            kwargs.update(dict(num_elements=num_elements))
        cls = Float if not is_list else FloatList
        return self.add(cls(**kwargs))

    def add_float(self, *args, **kwargs):
        """See `self.add_flt`."""
        return self.add_flt(*args, **kwargs)

    def add_pth(self,
                name: str,
                shortcut: str = None,
                help: str = None,
                required: bool = False,
                default: Any = None,
                must_exist: bool = False,
                is_dir: bool = False,
                is_list: bool = False,
                num_elements: int = None):
        """Specifies a path-type argument.

        Arguments:
            name (str): Name of the argument. Unlike `argparse`, preceding
                hyphens should be omitted. Allows specification by `--{name}`.
            shortcut (optional, str): Argument shortcut. Allows specification
                by `-{shortcut}`.
            help (optional, str): Help text.
            required (optional, str): Whether this argument is required.
            default (optional, *): Default value for the argument. Must be
                hashable (`__hash__`).
            must_exist (bool): The specified path must be an existing one.
                The argument parser will check whether path(s) do exist.
                (default: False)
            is_dir (bool): Whether the path is of directory type.
                The argument parser will check whether path(s) are directories.
                (default: False)
            is_list (bool): Whether this argument is of list type. If true,
                this argument can be specified with multiple values, i.e.

                    --{name} {val1} {val2} ... {valN}

                (default: False)
            num_elements (optional, int): Number of required elements.
                Applicable only if `is_list` is true.
        """
        kwargs = dict(
            name=name,
            shortcut=shortcut,
            help=help,
            required=required,
            default=default,
            must_exist=must_exist,
            is_dir=is_dir
        )
        if is_list:
            kwargs.update(dict(num_elements=num_elements))
        cls = Path if not is_list else PathList
        return self.add(cls(**kwargs))

    def add_path(self, *args, **kwargs):
        """See `self.add_pth`."""
        return self.add_pth(*args, **kwargs)

    def add_str(self,
                name: str,
                shortcut: str = None,
                help: str = None,
                required: bool = False,
                default: Any = None,
                min_length: int = None,
                max_length: int = None,
                regex: str = None,
                format: str = None,
                choices: Sequence[float] = None,
                is_list: bool = False,
                num_elements: int = None):
        """Specifies a string-type argument.

        Arguments:
            name (str): Name of the argument. Unlike `argparse`, preceding
                hyphens should be omitted. Allows specification by `--{name}`.
            shortcut (optional, str): Argument shortcut. Allows specification
                by `-{shortcut}`.
            help (optional, str): Help text.
            required (optional, str): Whether this argument is required.
            default (optional, *): Default value for the argument. Must be
                hashable (`__hash__`).
            min_length (optional, int): Minimum bound for length (inclusive).
            max_length (optional, int): Maximum bound for length (inclusive).
            regex (optional, str): Regular expression for matching values.
                The argument parser will check whether the value(s) conform
                the regular expression.
            format (optional, str): String format, as supported by JSONSchema.
                Some common formats include `date-time`, `time`, `url`, `email`.
                More information at JSONSchema reference.
            choices (optional, Sequence[str]): Enforces the argument to be
                one of these candidate strings.
            is_list (bool): Whether this argument is of list type. If true,
                this argument can be specified with multiple values, i.e.

                    --{name} {val1} {val2} ... {valN}

                (default: False)
            num_elements (optional, int): Number of required elements.
                Applicable only if `is_list` is true.
        """
        kwargs = dict(
            name=name,
            shortcut=shortcut,
            help=help,
            required=required,
            default=default,
            min_length=min_length,
            max_length=max_length,
            regex=regex,
            format=format,
            choices=choices
        )
        if is_list:
            kwargs.update(dict(num_elements=num_elements))
        cls = Str if not is_list else StrList
        return self.add(cls(**kwargs))

    def add_bol(self,
                name: str,
                shortcut: str = None,
                help: str = None,
                required: bool = False,
                invert: bool = False):
        """Specifies a bool-type argument. This is value-less argument, which is
        specified with only the keyword argument, i.e. `--{name}`.

        Arguments:
            name (str): Name of the argument. Unlike `argparse`, preceding
                hyphens should be omitted. Allows specification by `--{name}`.
            shortcut (optional, str): Argument shortcut. Allows specification
                by `-{shortcut}`.
            help (optional, str): Help text.
            required (optional, str): Whether this argument is required.
            invert (bool): Whether to invert bool operation. The default
                behavior stores True when the argument is specified and False
                for otherwise.
                (default: False)
        """
        return self.add(Bool(
            name=name,
            shortcut=shortcut,
            help=help,
            required=required,
            invert=invert
        ))

    def add_bool(self, *args, **kwargs):
        """See `self.add_bol`"""
        return self.add_bol(*args, **kwargs)

    def add_intlist(self, *args, **kwargs):
        return self.add_int(*args, **kwargs, is_list=True)

    def add_floatlist(self, *args, **kwargs):
        return self.add_flt(*args, **kwargs, is_list=True)

    def add_pathlist(self, *args, **kwargs):
        return self.add_pth(*args, **kwargs, is_list=True)

    def add_strlist(self, *args, **kwargs):
        return self.add_str(*args, **kwargs, is_list=True)

    def schema(self) -> dict:
        """Returns this argument specification in terms of JSONSchema."""
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "$id": str(hash((self.name, self.desc))),
            "type": "object",
            "properties": {arg.name: arg.json_schema()
                           for arg in self._args.values()},
            "required": [arg.name for arg in
                         self._args.values() if arg.required]
        }

    def validate(self, args: dict):
        """Validate a dictionary of arguments against this argument spec."""
        jsonschema.validate(args, self.schema())

    def print_schema(self):
        schema = self.schema()
        sys.stdout.write(json.dumps(schema, ensure_ascii=False, indent=2))

    def _parse_config(self, config: dict):
        args = []
        for k, v in config.items():
            if v is None:
                continue
            arg: Argument = self._args[k]
            if isinstance(v, bool):
                assert isinstance(arg, Bool)
                if v != arg.invert:
                    args.append(f"--{k}")
            elif isinstance(v, str):
                args.extend([f"--{k}", v])
            elif isinstance(v, list):
                args.append(f"--{k}")
                args.extend(map(str, v))
            else:
                args.extend((f"--{k}", str(v)))
        return args

    def _argparse(self, args, namespace=None):
        if self.on_extra == "ignore" or self.on_extra == "warn":
            namespace, extra_args = \
                self._parser.parse_known_args(args, namespace)
            if self.on_extra == "warn":
                warnings.warn(f"following arguments are not argparsed: "
                              f"{extra_args}")
        elif self.on_extra == "error":
            namespace = self._parser.parse_args(args, namespace)
        return namespace

    def parse(self, args: Optional[Sequence[str]] = None,
              namespace: Optional[argparse.Namespace] = None
              ) -> argparse.Namespace:
        if not self.reserved_arg_names:
            return self._argparse(args=args, namespace=namespace)
        reserved_namespace, _ = self._reserved_parser.parse_known_args(args)
        if self.add_schema_arg and getattr(reserved_namespace,
                                           self.schema_arg_name):
            self.print_schema()
            exit(0)
        if self.add_config_arg:
            config_path = getattr(reserved_namespace, self.config_arg_name)
            if config_path is not None:
                with open(config_path, "r") as f:
                    config = yaml.safe_load(f)
                self.validate(config)
                config_args = self._parse_config(config)
                args = list(config_args) + list(args or sys.argv[1:])
        namespace = self._argparse(args=args, namespace=namespace)
        return namespace



__all__ = ["private_field"]

from dataclasses import field, MISSING


def private_field(default=MISSING, default_factory=MISSING):
    return field(init=False, repr=False, compare=False,
                 default=default, default_factory=default_factory)



__all__ = ["Argument", "ValueArgument", "ListArgument",
           "Bool", "Str", "Path", "Int", "Float",
           "StrList", "PathList", "IntList", "FloatList", "Yaap"]
__version__ = "0.4"

from .argument import *
from .yaap import *



from setuptools import setup

setup(
    name="yaap",
    author="Kang Min Yoo",
    author_email="kangmin.yoo@gmail.com",
    description="Yet another argument parser (Yaap)",
    url="https://github.com/kaniblu/yaap",
    keywords="argument parser config file",
    version="0.4",
    packages=["yaap"],
    install_requires=list(s.rstrip("\n") for s in open("requirements.txt", "r"))
)



__all__ = ["Argument", "ValueArgument", "ListArgument",
           "Bool", "Str", "Path", "Int", "Float",
           "StrList", "PathList", "IntList", "FloatList"]

import re
import os
import argparse
from dataclasses import dataclass
from typing import List, Dict, Tuple, Any, Sequence


@dataclass(frozen=True)
class Argument:
    name: str
    shortcut: str = None
    help: str = None
    required: bool = False

    @property
    def default_value(self):
        return

    def json_schema(self) -> dict:
        schema = {"title": self.name}
        if self.help:
            schema["description"] = self.help
        if self.default_value:
            schema["default"] = self.default_value
        return schema

    def generate_args(self) -> Tuple[List, Dict]:
        args = list()
        if self.shortcut is not None:
            if len(self.shortcut) > 1:
                raise ValueError(f"argument shortcut must be a character: "
                                 f"len({self.shortcut}) > 1")
            if not self.shortcut.isalpha():
                raise ValueError(f"argument shortcut must be an alphabet: "
                                 f"{self.shortcut}")
            args.append(f"-{self.shortcut}")
        args.append(f"--{self.name}")
        kwargs = dict()
        kwargs["help"] = self.help
        kwargs["required"] = self.required
        return args, kwargs


@dataclass(frozen=True)
class Bool(Argument):
    invert: bool = False

    @property
    def default_value(self):
        return self.invert

    def json_schema(self) -> dict:
        return dict(type="boolean", **super().json_schema())

    def generate_args(self) -> Tuple[List, Dict]:
        args, kwargs = super(Bool, self).generate_args()
        kwargs["action"] = "store_true" if not self.invert else "store_false"
        kwargs["default"] = self.default_value
        if "required" in kwargs:
            del kwargs["required"]
        default_msg = "(default: False)"
        if "help" in kwargs and kwargs["help"]:
            kwargs["help"] = f"{kwargs['help']} {default_msg}"
        else:
            kwargs["help"] = default_msg
        return args, kwargs


@dataclass(frozen=True)
class ValueArgument(Argument):
    default: Any = None

    @property
    def default_value(self):
        if self.default is not None:
            return self.cast(self.default)

    def validate(self, value) -> bool:
        raise NotImplementedError

    def cast(self, value):
        return value

    def type_cast(self, value):
        if value is None:
            return
        value = self.cast(value)
        if not self.validate(value):
            raise argparse.ArgumentTypeError(f"validation failed: {value}")
        return value

    def generate_args(self) -> Tuple[List, Dict]:
        args, kwargs = super(ValueArgument, self).generate_args()
        kwargs["type"] = self.type_cast
        if self.default is not None:
            kwargs["default"] = self.default_value
            if "help" in kwargs and kwargs["help"]:
                kwargs["help"] = f"{kwargs['help']} (default: %(default)s)"
            else:
                kwargs["help"] = "(default: %(default)s)"
        return args, kwargs


@dataclass(frozen=True)
class Str(ValueArgument):
    choices: Sequence[str] = None
    min_length: int = None
    max_length: int = None
    regex: str = None
    format: str = None

    def cast(self, value):
        return str(value)

    def json_schema(self) -> dict:
        schema = super(Str, self).json_schema()
        schema["type"] = ["string"]
        if not self.required:
            schema["type"].append("null")
        if self.min_length is not None:
            schema["minLength"] = self.min_length
        if self.max_length is not None:
            schema["maxLength"] = self.max_length
        if self.regex is not None:
            schema["pattern"] = self.regex
        if self.format is not None:
            schema["format"] = self.format
        if self.choices is not None:
            schema["enum"] = list(self.choices)
        return schema

    def validate(self, value: str) -> bool:
        if self.regex is not None:
            return re.match(self.regex, value) is not None
        if self.min_length is not None and len(value) < self.min_length:
            return False
        if self.max_length is not None and len(value) > self.max_length:
            return False
        return True

    def generate_args(self) -> Tuple[List, Dict]:
        args, kwargs = super(Str, self).generate_args()
        if self.choices is not None:
            kwargs["choices"] = list(self.choices)
        return args, kwargs


@dataclass(frozen=True)
class Path(ValueArgument):
    must_exist: bool = False
    is_dir: bool = False

    def cast(self, value):
        return os.path.abspath(os.path.realpath(str(value)))

    def validate(self, value) -> bool:
        if self.must_exist and self.is_dir and not os.path.isdir(value):
            return False
        if self.must_exist and not os.path.exists(value):
            return False
        return True

    def json_schema(self) -> dict:
        schema = super(Path, self).json_schema()
        schema["type"] = ["string"]
        if not self.required:
            schema["type"].append("null")
        schema["format"] = "uri"
        return schema

    def generate_args(self) -> Tuple[List, Dict]:
        return super(Path, self).generate_args()


@dataclass(frozen=True)
class Int(ValueArgument):
    min_bound: int = None
    max_bound: int = None
    multiples: int = None
    choices: Sequence[int] = None

    def json_schema(self) -> dict:
        schema = super(Int, self).json_schema()
        schema["type"] = ["integer"]
        if not self.required:
            schema["type"].append("null")
        if self.multiples is not None:
            schema["multipleOf"] = self.multiples
        if self.min_bound is not None:
            schema["minimum"] = self.min_bound
        if self.max_bound is not None:
            schema["exclusiveMaximum"] = self.max_bound
        return schema

    def cast(self, value):
        return int(value)

    def validate(self, value: int) -> bool:
        if self.min_bound is not None and value < self.min_bound:
            return False
        if self.max_bound is not None and value >= self.max_bound:
            return False
        if self.choices is not None and value not in self.choices:
            return False
        if self.multiples is not None and value % self.multiples != 0:
            return False
        return True


@dataclass(frozen=True)
class Float(ValueArgument):
    min_bound: float = None
    max_bound: float = None
    multiples: float = None
    choices: Sequence[float] = None

    def json_schema(self) -> dict:
        schema = super(Float, self).json_schema()
        schema["type"] = ["number"]
        if not self.required:
            schema["type"].append("null")
        if self.multiples is not None:
            schema["multipleOf"] = self.multiples
        if self.min_bound is not None:
            schema["minimum"] = self.min_bound
        if self.max_bound is not None:
            schema["exclusiveMaximum"] = self.max_bound
        return schema

    def cast(self, value):
        return float(value)

    def validate(self, value) -> bool:
        if self.min_bound is not None and value < self.min_bound:
            return False
        if self.max_bound is not None and value >= self.max_bound:
            return False
        if self.choices is not None and value not in self.choices:
            return False
        if self.multiples is not None and value % self.multiples != 0:
            return False
        return True


@dataclass(frozen=True)
class ListArgument(ValueArgument):
    num_elements: int = None

    @property
    def default_value(self):
        if self.default is not None:
            return list(self.default)

    def json_schema(self) -> dict:
        schema = super(ValueArgument, self).json_schema()
        schema["type"] = ["array"]
        if not self.required:
            schema["type"].append("null")
        schema["items"] = super(ListArgument, self).json_schema()
        schema["items"]["title"] = schema["items"]["title"] + "/item"
        return schema

    def generate_args(self) -> Tuple[List, Dict]:
        args, kwargs = super(ListArgument, self).generate_args()
        if self.num_elements is None:
            kwargs["nargs"] = "+" if self.required else "*"
        else:
            kwargs["nargs"] = self.num_elements
        kwargs["default"] = self.default_value
        return args, kwargs

    def validate(self, value) -> bool:
        return super(ListArgument, self).validate(value)


@dataclass(frozen=True)
class StrList(ListArgument, Str):
    pass


@dataclass(frozen=True)
class IntList(ListArgument, Int):
    pass


@dataclass(frozen=True)
class FloatList(ListArgument, Float):
    pass


@dataclass(frozen=True)
class PathList(ListArgument, Path):
    pass